(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/RL.sml,v
   Comment: Language of restriction commands for specializing statecharts
   Release: scope-0-15_final 

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   RL.sml,v 1.7 2003/05/05 14:04:58 wasowski Exp
*)

structure RL :> RL = struct 

open Err

structure SE = SymbolEnv
structure S = Symbol

type 'a env = 'a SE.env and symbol = S.symbol 
type event = symbol and set = Set.set


(* abstract syntax used in parsing *)

datatype property = NeverE of string
		  | ConstrV of Types.simplety * string * Types.value
		  | ConstrF of Types.simplety * string * Types.value
		  | Dead of Types.simplety * string
		  | Pure of Types.simplety * string

type restriction = string * string list * property list 

datatype block = Inter of restriction
	       | IBinding of string * string

type rl_ast = block list

(* refined representation forwarded to specializer *)
(* all references fully expanded                   *)

type vaprop = Types.simplety * Types.value option * bool 
		(* type, value, live *)

type fuprop = Types.simplety * Types.value option * bool
		(* retty, value, pure *)

type restr = { 
		vaprop :vaprop env,
		fuprop :fuprop env,
		nevere :set
	      }

type rl = {
		restr :restr env,
		sys   :restr env
	}

val empty_restr :restr = {vaprop=SE.empty, fuprop=SE.empty, nevere=Set.empty } 
val empty = {restr=SE.empty, sys= SE.empty}

(*** helper operations on restrictions and specs ***)

fun get_restr (rl :rl) name :restr = 
	SE.lookup (#restr rl) name	(* raises NotFound *)

fun get_sys (rl :rl) name :restr =
	SE.lookup (#sys rl) name	(* raises NotFound *)

fun defd_sys (rl :rl) name :bool = 
	SE.defined (#sys rl) name

fun defd_restr (rl :rl) name :bool = 
	SE.defined (#restr rl) name

fun bind_sys (rl :rl) (name,restr) :rl = 
	{sys= SE.bind1 (#sys rl) (name,restr), restr= #restr rl}

fun bind_restr (rl :rl) (name,restr) :rl = 
	{sys= #sys rl, restr= SE.bind1 (#restr rl) (name,restr)}

fun add_nevere ({vaprop,fuprop,nevere} :restr) (e :event) =
	{vaprop= vaprop, fuprop= fuprop, nevere= Set.add(nevere,e)}

fun add_fuprop ({vaprop,fuprop,nevere} :restr) (f :symbol,p :fuprop) =
	{vaprop= vaprop, fuprop= SE.bind1 fuprop (f,p), nevere= nevere}

fun add_vaprop ({vaprop,fuprop,nevere} :restr) (v :symbol,p :fuprop) =
	{vaprop= SE.bind1 vaprop (v,p), fuprop= fuprop, nevere= nevere}

(*** flattening of rl specifications ***)

(* overwrite variable properties *)

fun typematch (t1,t2) = 
	if t1<>t2 then semErr "Type mismatch in overloading" else t1 

fun valuematch (NONE,  SOME v) = SOME v
  | valuematch (SOME x,SOME y) = 
	if x = y then SOME x 
	else semErr "Attempt to overload variable constraint with new value"
  | valuematch (v1,v2)         = v1

fun plusv ((t1,v1,l1),(t2,v2,l2)) :vaprop = 
	let val t = typematch (t1,t2) before "karol ma kooooooota" 
	    val v = valuematch (v1,v2)
	    val l = l1 andalso l2
	in (t,v,l) end 

fun plusf ((t1,v1,p1),(t2,v2,p2)) :fuprop =
       let val t = typematch (t1,t2)
           val v = valuematch(v1,v2)
           val p = p1 orelse p2
       in (t,v,p) end

fun  add_restr (restr :restr,anc :restr) :restr =
	{ vaprop= SE.join plusv (#vaprop restr,#vaprop anc),
	  fuprop= SE.join plusf (#fuprop restr,#fuprop anc),
	  nevere= Set.union (#nevere restr, #nevere anc) }
	


fun add_prop (prop :property, restr: restr) :restr =
	let fun add_fu (n,t,v,p) =
		let val sn = S.symbol n
		    val p1 = plusf (SE.lookup (#fuprop restr) sn,(t,v,p)) 
			     handle SE.NotFound => (t,v,p)
		in add_fuprop restr (sn,p1) end
		
	    fun add_va (n,t,v,l) =
		let val sn = S.symbol n
		    val p1 = plusv (SE.lookup (#vaprop restr) sn,(t,v,l))
			     handle SE.NotFound => (t,v,l)
		in add_vaprop restr (sn,p1) end
	in	
		case prop of
			  NeverE e	    => add_nevere restr (S.symbol e)
			| ConstrV (t,nam,v) => add_va (nam,t,SOME v,true )
			| ConstrF (t,nam,v) => add_fu (nam,t,SOME v,false)
			| Dead (t,nam)	    => add_va (nam,t,NONE,  false)
			| Pure (t,nam)      => add_fu (nam,t,NONE,  true )
	end 
				
			
			  

fun flatten1 (bl :block,rl :rl) = 
	case bl of
		  Inter (nam,ancs,props) => 
			let val snam = S.symbol nam
			    val sancs = map S.symbol ancs
			    val _ = if defd_restr rl snam then semErr
					("Redefinition of restriction " ^ nam)
				    else ()
			    val bancs = map (get_restr rl) sancs
					handle SE.NotFound => semErr 
					"Restriction of undefined interface " 

			    val body = foldl add_restr empty_restr bancs
			    val body1 = foldl add_prop body props
			in bind_restr rl (snam,body1) end 

		| IBinding (restr,sys) =>
			let val ssys = S.symbol sys
			    val srestr = S.symbol restr
			    val _ = if defd_sys rl ssys then semErr 
					("Redefinition of restriction \
					 \for system " ^ sys)
				    else ()
			    val _ = if defd_restr rl srestr then ()
				    else semErr (
					"Reference to undefined restriction "
					^ restr)
			in bind_sys rl (ssys,get_restr rl srestr) end 



fun flatten (ast :rl_ast, rl:rl) :rl =
	foldl flatten1 rl ast 
	


	
(*
	type rlspec = rl list

	(* rlcheck should discover as many errors as possible,
	   not stop after the first one. It may be seen as a test
	   procedure. Probably in final implementation correction
	   will be checked at IL level. *)

	fun rlch (sys :system) (rl :rl) :bool = 
		case rl of
			(* only external events can be removed currently *)
			RmEvent e => not (#1 (SymbolEnv.lookup 
						(#evtypes sys) e))
			| _ => true

	fun rlcheck1 (sys :system) (rls :rlspec) :bool =
		foldl (fn (rl,e) => rlch sys rl andalso e) true rls
		

	fun rl_check rlspecs model :bool =
		ListPair.foldl (fn (sys,rls,e) => rlcheck1 sys rls andalso e)
			true (model,rlspecs)
*)
end
