(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/lr.sml,v
   Comment: intermediate representation (used by CodGenC1 & CodGenCF).
   Release: scope-0-15_final 
  
     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   lr.sml,v 1.6 2003/07/21 23:01:22 wasowski Exp
*)


structure LR = struct

open CodGen Types Exp Err

structure SE = SymbolEnv
structure S  = Symbol

infix  3 --

type symbol = Symbol.symbol
type 'a env = 'a SE.env
type ''a ulist = ''a UList.ulist

(* runtime data layout markers (aka opcodes) *)

datatype marker = STMRK | HMRK | IMRK | FPSCOP | NFPSCOP 

(* low level representation byte code *)

datatype lratom = 
	  LRMrk          of marker	(* bytecode marker                  *)
	| LRAndStateRef  of symbol	(* reference to andstate by name    *)
        | LROrStateRef   of symbol	(* reference to orstate  by name    *)
	| LRNoEvent			(* a special "no event" event       *)
	| LREventRef     of symbol	(* reference to trans group by event*)
	| LRHistoryRef   of int		(* reference to a cell in history v.*)
	| LREnEx         of int * int	(* reference to entry/exit actions  *) 
	| LRAcGd         of int * int	(* reference to guard and action    *)
	| LRCondCounters of int * int	(* counter of pos and neg conds     *)
	| LRStateLabel   of symbol	(* mark start of state,no code emit *)
	| LREventLabel   of symbol	(* start of group of trans, no code *)

(* transitions *)	

type    tenv = lratom list env		(* group transitions by trigger        *)
type tranidx = int env			(* map events to groups of transitions *)

(* events *)

type eenv = int env

(* variables *)

type vartype = Types.vartype
type vartypes= vartype env
type venv    = value env


(* events and signals *)

type signals = (lratom list) ulist


type evtypes = evtype env

val  sigsize = length


(* actions and guards *)

val  guardsize = fn _ => 1
type guard  = bexp list		(* unchanged from Absyn but deliberately      *)
				(* redefined                                  *)
type guards = guard ulist	(* where guards is a raw source guard list    *)
				(* from Absyn.sml                             *)
type guardref = int		
type guardrefs = guardref list

val always     :guard    = [Cst (CstI 1)]


val  actionsize = fn _ => 1

type sexp = symbol * exp list		(* deliberately copied from Absyn.sml *)
					(* name, [param]                      *)
                                       
type action = command list * sexp list	(* copied from Absyn.sml, but         *)
                                        (* deliberately redefined             *)
                                        (* to make LR independent of Absyn    *)
   
type actions = action ulist	(* where actions is a raw source action 
				   from Absyn.sml *)

val null_action :action = ([],[])

(* type information about the model *)

type lrsizes = {	
	
	actionref :int,		(* inferred type of action refs              *)
	guardref  :int,		(* inferred type of guard ref type           *)
	orstref   :int,		(* inferred type of orstate (machine) refs   *)
	andstref  :int,		(* inferred type of andstate (basic st) refs *)
	signalref :int,		(* inferred return type for all acts         *)
	evnumber  :int		(* events amount: id of last event+1         *)

}



(* split an integer number data into younger part of bytes length
   and older part - what is remaining. I think I should have put more effort
   in following packing functions. What I have now is big, slow and ugly. *)

local fun f 0 res = res -1
	| f n res = f (n-1) (256 * res)
in fun ones n = f n 1 end

fun partition_bits (data :int) (bytes :int) :int * int =
	let val mask   = Word.fromInt (ones bytes)
	    val wdata  = Word.fromInt data
	    val suffix = Word.andb(wdata,mask)
	    val prefix = Word.>>(wdata,Word.fromInt (bytes * 8))
	in (Word.toInt prefix, Word.toInt suffix) end;

fun dist chunk data 0 res = res
  | dist chunk data 1 res = data::res
  | dist chunk data n res = let val (pre,suf) = partition_bits chunk data 
			    in dist chunk pre (n-1) (suf::res) end

(* generic packing of integers to arrays of arbitrary size target chunks. *)

fun pack1 (data :int) (src_size :int) (tgt_size :int) :int list =
	let val chunks = Real.ceil((Real.fromInt src_size / Real.fromInt tgt_size))
	in dist tgt_size data chunks [] end;

fun pack2 (data1, data2) (src1,src2) (tgt_size) :int list = 
	if src1+src2 <= tgt_size then [(ones src1 + 1) * data2 + data1]
	else let val c1 = Real.ceil((Real.fromInt src1/Real.fromInt tgt_size))
		 val c2 = Real.ceil((Real.fromInt src2/Real.fromInt tgt_size))
		 val d2 = dist tgt_size data2 c2 []
		 val d1 = dist tgt_size data1 c1 d2
	in d1 end;

(* generic size estimation of packed data. Return number of target chunks *)

fun measure1 (src_size :int) (tgt_size :int) :int =
	Real.ceil((Real.fromInt src_size / Real.fromInt tgt_size)) 

fun measure2 (src1,src2) (tgt_size :int) :int =
	if src1+src2 <= tgt_size then 1
	else Real.ceil((Real.fromInt src1/Real.fromInt tgt_size)) +
	     Real.ceil((Real.fromInt src2/Real.fromInt tgt_size))

(* generation of generic macro accessing packed data. *)

(* some generic operations about packing and distributing integers over
   various integer sizes. USed in details of scheduling and compilation *)


fun intgr (var :string) (cells :int) :codgen = 
	let fun f 0 = eoc
	      | f 1 = sg "+ (*((ptr)++))"
	      | f n = sg "+ (*((ptr)++)) << (sizeof(*(ptr)) * " 
				-- ig (n-1) -- sg ")" -- f (n-1)
	in cg [ "{ (", var, ") = " ] -- f cells -- sg "; }" end

(* only creates the body of macro. parameter is supposed to be called d *)

(*
fun access1 (src_size :int) (tgt_size :int) :codgen =
	let val 
*)


fun access2 (src1 :int, src2 :int) (tgt_size :int) :codgen =
	if src1+src2 <= tgt_size then
		cg ["{ (i) = *(ptr) & ", (Int.toString (ones src1)), "; ",
		    "j = *((ptr)++) >>  ", (Int.toString (src1 * 8)), "; }"]
	else let val c1 = Real.ceil((Real.fromInt src1/Real.fromInt tgt_size))
		 val c2 = Real.ceil((Real.fromInt src2/Real.fromInt tgt_size))
		 val m1 = intgr "i" c1 
		 val m2 = intgr "j" c2
	in sg "{ " -- m1 -- m2 -- sg " }" end


(*
fun integr src_size tgt_size macro :codgen =
	if src_size >= tgt_size then fn os => "(d)"::os
	else let val mask = ones src_size
	     in 
		integr src_size (tgt_size - src_size) macro 
			o (fn os => " & d) + " ::itoa mask ::"(("::os) 
	     end
*)	

(* type of textual substitutions *)

type subst = symbol env

(* substitution application *)

fun subst_exp (subst :subst) (e :exp) :exp = 
	case e of Access a => Access (subst_acc subst a)
		| Prim1(ope,e) => Prim1(ope,subst_exp subst e)
		| Prim2(ope,e1,e2) => 
			Prim2(ope, subst_exp subst e1, subst_exp subst e2)
		| Call oe => Call (subst_oexp subst oe)
		| anything => anything

and subst_acc (subst :subst) (a :access) :access =
	case a of AccessVar v => 
			AccessVar (SE.lookup subst v handle NotFound => v)
		| AccessArr (a,e) =>
			AccessArr (SE.lookup subst a handle NotFound => a,
				   subst_exp subst e)

and subst_oexp (subst :subst) (f,ps) :oexp = (f,map (subst_exp subst) ps)
	
fun subst_sexp (subst :subst) (e,ps) :sexp = (e, map (subst_exp subst) ps)

fun subst_cmd (subst :subst) cmd = 
	case cmd of
		  Expr e => Expr (subst_exp subst e)
		| Assgn (a,e) => Assgn (subst_acc subst a,subst_exp subst e)

fun subst_action (subst :subst) ((cmds,sigs) :action) :action =
	(map (subst_cmd subst) cmds, map (subst_sexp subst) sigs)

fun subst_guard (subst :subst) guard :guard =
	map (subst_exp subst) guard

(* compute substitution based on transition's trigger *)

fun comp_subst (ets :evtype env) (e,ps) :symbol env =
	(case SE.lookup ets e of 
	  (true,_) => 
		SE.empty before
		(if null ps then () else warning ("Parameters for signal "
						^ Symbol.name e ^ "ignored"))
	| (false,[]) => SE.empty
	| (false,ts) =>
		let val uf = "CurrEvent.fields._E_" ^ Symbol.name e ^ ".f"
		    val fields = List.tabulate (length ts,
				fn i => Symbol.symbol(uf^ Int.toString i))
		in SE.bindZip SE.empty (ps,fields)  end
	)handle NotFound => impossible "#20020709c"

end



